/*
 * Helper functions for multiple mount protection(MMP).
 *
 * Copyright (C) 2006, 2007 by Kalpak Shah <kalpak@clusterfs.com>
 *
 * %Begin-Header%
 * This file may be redistributed under the terms of the GNU Public
 * License.
 * %End-Header%
 */

#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/time.h>

#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2fs.h"

errcode_t ext2fs_mmp_read(ext2_filsys fs, blk_t mmp_blk, char *buf)
{
	struct mmp_struct *mmp_s;
	errcode_t retval;

	if ((mmp_blk < fs->super->s_first_data_block) ||
	    (mmp_blk >= fs->super->s_blocks_count))
		return EXT2_ET_MMP_BAD_BLOCK;

	/*
	 * Make sure that we read direct from disk by reading only
	 * sizeof(stuct mmp_struct) bytes.
	 */
	retval = io_channel_read_blk(fs->io, mmp_blk,
				     -(int)sizeof(struct mmp_struct), buf);
	if (retval)
		return retval;

	mmp_s = (struct mmp_struct *) buf;

#ifdef EXT2FS_ENABLE_SWAPFS
	if (fs->flags & EXT2_FLAG_SWAP_BYTES)
		ext2fs_swap_mmp(mmp_s);
#endif

	if (mmp_s->mmp_magic != EXT2_MMP_MAGIC)
		return EXT2_ET_MMP_MAGIC_INVALID;

	return 0;
}

errcode_t ext2fs_mmp_write(ext2_filsys fs, blk_t mmp_blk, char *buf)
{
	struct mmp_struct *mmp_s = (struct mmp_struct *) buf;
	struct timeval tv;
	int retval;

	gettimeofday(&tv, 0);
	mmp_s->mmp_time = tv.tv_sec;

#ifdef EXT2FS_ENABLE_SWAPFS
	if (fs->super->s_magic == ext2fs_swab16(EXT2_SUPER_MAGIC))
		ext2fs_swap_mmp(mmp_s);
#endif

	retval = io_channel_write_blk(fs->io, mmp_blk,
				      -(int)sizeof(struct mmp_struct), buf);

#ifdef EXT2FS_ENABLE_SWAPFS
	if (fs->super->s_magic == ext2fs_swab16(EXT2_SUPER_MAGIC))
		ext2fs_swap_mmp(mmp_s);
#endif

	/*
	 * Make sure the block gets to disk quickly.
	 */
	io_channel_flush(fs->io);
	return retval;
}

long int ext2fs_mmp_new_seq()
{
	long int new_seq;

	do {
		new_seq = random();
	} while (new_seq > EXT2_MMP_SEQ_MAX);

	return new_seq;
}

errcode_t ext2fs_mmp_init(ext2_filsys fs)
{
	struct ext2_super_block *sb = fs->super;
	struct mmp_struct *mmp_s = NULL;
	blk_t mmp_block;
	int error;

	error = ext2fs_read_bitmaps(fs);
	if (error)
		goto out;

	error = ext2fs_new_block(fs, 0, 0, &mmp_block);
	if (error)
		goto out;

	ext2fs_block_alloc_stats(fs, mmp_block, +1);
	sb->s_mmp_block = mmp_block;

	if (fs->mmp_buf == NULL) {
		error = ext2fs_get_mem(fs->blocksize, &fs->mmp_buf);
		if (error)
			goto out;
	}

	mmp_s = (struct mmp_struct *)fs->mmp_buf;
	memset(mmp_s, 0, sizeof(struct mmp_struct));

	mmp_s->mmp_magic = EXT2_MMP_MAGIC;
	mmp_s->mmp_seq = EXT2_MMP_SEQ_CLEAN;
	mmp_s->mmp_time = 0;
#if _BSD_SOURCE || _XOPEN_SOURCE >= 500
	gethostname(mmp_s->mmp_nodename, sizeof(mmp_s->mmp_nodename));
#else
	mmp_s->mmp_nodename[0] = '\0';
#endif
	strncpy(mmp_s->mmp_bdevname, fs->device_name,
		sizeof(mmp_s->mmp_bdevname));

	mmp_s->mmp_check_interval = EXT2_MMP_MIN_CHECK_INTERVAL;

	error = ext2fs_mmp_write(fs, mmp_block, fs->mmp_buf);
	if (error) {
		ext2fs_free_mem(&fs->mmp_buf);
		fs->mmp_buf = NULL;
		goto out;
	}

	sb->s_mmp_update_interval = EXT2_MMP_UPDATE_INTERVAL;

out:
	return error;
}

/*
 * Make sure that the fs is not mounted or being fsck'ed while opening the fs.
 */
errcode_t ext2fs_mmp_start(ext2_filsys fs)
{
	blk_t mmp_blk = fs->super->s_mmp_block;
	struct mmp_struct *mmp_s;
	unsigned seq;
	unsigned int mmp_check_interval;
	errcode_t retval = 0;

	if (fs->mmp_buf == NULL) {
		retval = ext2fs_get_mem(fs->blocksize, &fs->mmp_buf);
		if (retval)
			goto mmp_error;
	}

	retval = ext2fs_mmp_read(fs, mmp_blk, fs->mmp_buf);
	if (retval)
		goto mmp_error;

	mmp_s = (struct mmp_struct *)fs->mmp_buf;

	mmp_check_interval = fs->super->s_mmp_update_interval;
	if (mmp_check_interval < EXT2_MMP_MIN_CHECK_INTERVAL)
		mmp_check_interval = EXT2_MMP_MIN_CHECK_INTERVAL;

	/*
	 * If check_interval in MMP block is larger, use that instead of
	 * check_interval from the superblock.
	 */
	if (mmp_s->mmp_check_interval > mmp_check_interval)
		mmp_check_interval = mmp_s->mmp_check_interval;

	seq = mmp_s->mmp_seq;
	if (seq == EXT2_MMP_SEQ_CLEAN)
		goto clean_seq;
	if (seq == EXT2_MMP_SEQ_FSCK) {
		retval = EXT2_ET_MMP_FSCK_ON;
		goto mmp_error;
	}

	if (seq > EXT2_MMP_SEQ_FSCK) {
		retval = EXT2_ET_MMP_UNKNOWN_SEQ;
		goto mmp_error;
	}

	sleep(2 * mmp_check_interval + 1);

	retval = ext2fs_mmp_read(fs, mmp_blk, fs->mmp_buf);
	if (retval)
		goto mmp_error;

	if (seq != mmp_s->mmp_seq) {
		retval = EXT2_ET_MMP_FAILED;
		goto mmp_error;
	}

clean_seq:
	mmp_s->mmp_seq = seq = ext2fs_mmp_new_seq();
#if _BSD_SOURCE || _XOPEN_SOURCE >= 500
	gethostname(mmp_s->mmp_nodename, sizeof(mmp_s->mmp_nodename));
#else
	strcpy(mmp_s->mmp_nodename, "unknown host");
#endif
	strncpy(mmp_s->mmp_bdevname, fs->device_name,
		sizeof(mmp_s->mmp_bdevname));

	retval = ext2fs_mmp_write(fs, mmp_blk, fs->mmp_buf);
	if (retval)
		goto mmp_error;

	sleep(2 * mmp_check_interval + 1);

	retval = ext2fs_mmp_read(fs, mmp_blk, fs->mmp_buf);
	if (retval)
		goto mmp_error;

	if (seq != mmp_s->mmp_seq) {
		retval = EXT2_ET_MMP_FAILED;
		goto mmp_error;
	}

	mmp_s->mmp_seq = EXT2_MMP_SEQ_FSCK;
	retval = ext2fs_mmp_write(fs, mmp_blk, fs->mmp_buf);
	if (retval)
		goto mmp_error;

	return 0;

mmp_error:
	return retval;
}

/*
 * Clear the MMP usage in the filesystem.  If this function returns an
 * error EXT2_ET_MMP_CHANGE_ABORT it means the filesystem was modified
 * by some other process while in use, and changes should be dropped, or
 * risk filesystem corruption.
 */
errcode_t ext2fs_mmp_stop(ext2_filsys fs)
{
	blk_t mmp_blk = fs->super->s_mmp_block;
	struct mmp_struct *mmp, *mmp_cmp;
	errcode_t retval;

	if (fs->mmp_cmp == NULL) {
		retval = ext2fs_get_mem(fs->blocksize, &fs->mmp_cmp);
		if (retval)
			goto mmp_error;
	}

	retval = ext2fs_mmp_read(fs, mmp_blk, fs->mmp_cmp);
	if (retval)
		goto mmp_error;

	/*
	 * This is important since we may come here just after when MMP feature
	 * is set and fs->mmp_buf is NULL
	 */
	if (fs->mmp_buf == NULL)
		goto check_skipped;

	/*
	 * Make sure that the MMP block is not changed.
	 */
	mmp = (struct mmp_struct *)fs->mmp_buf;
	mmp_cmp = (struct mmp_struct *)fs->mmp_cmp;
	if (memcmp(mmp, mmp_cmp, sizeof(struct mmp_struct)))
		return EXT2_ET_MMP_CHANGE_ABORT;

check_skipped:
	mmp_cmp->mmp_seq = EXT2_MMP_SEQ_CLEAN;
	retval = ext2fs_mmp_write(fs, mmp_blk, fs->mmp_cmp);

mmp_error:
	if (fs->mmp_buf) {
		ext2fs_free_mem(&fs->mmp_buf);
		fs->mmp_buf = NULL;
	}
	if (fs->mmp_cmp) {
		ext2fs_free_mem(&fs->mmp_cmp);
		fs->mmp_cmp = NULL;
	}

	return retval;
}

#define EXT2_MIN_MMP_UPDATE_INTERVAL 60

/*
 * Update the on-disk mmp buffer, after checking that it hasn't been changed.
 */
errcode_t ext2fs_mmp_update(ext2_filsys fs)
{
	blk_t mmp_blk = fs->super->s_mmp_block;
	struct mmp_struct *mmp, *mmp_cmp;
	struct timeval tv;
	errcode_t retval;

	if (!(fs->super->s_feature_incompat & EXT4_FEATURE_INCOMPAT_MMP) ||
	    !(fs->flags & EXT2_FLAG_RW) || (fs->flags & EXT2_FLAG_SKIP_MMP))
		return 0;

	gettimeofday(&tv, 0);
	if (tv.tv_sec - fs->mmp_last_written < EXT2_MIN_MMP_UPDATE_INTERVAL)
		return 0;

	if (fs->mmp_cmp == NULL) {
		retval = ext2fs_get_mem(fs->blocksize, &fs->mmp_cmp);
		if (retval)
			goto mmp_error;
	}

	retval = ext2fs_mmp_read(fs, mmp_blk, fs->mmp_cmp);
	if (retval)
		goto mmp_error;

	mmp = (struct mmp_struct *)fs->mmp_buf;
	mmp_cmp = (struct mmp_struct *)fs->mmp_cmp;

	if (memcmp(mmp, mmp_cmp, sizeof(struct mmp_struct)))
		return EXT2_ET_MMP_CHANGE_ABORT;

	mmp->mmp_time = tv.tv_sec;
	fs->mmp_last_written = tv.tv_sec;
	mmp->mmp_seq = EXT2_MMP_SEQ_FSCK;
	retval = ext2fs_mmp_write(fs, mmp_blk, fs->mmp_buf);

mmp_error:
	return retval;
}
