#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
extern char *optarg;
extern int optind;
#endif

#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2fs.h"
#include "e2freefrag.h"

void usage(const char *prog)
{
	fprintf(stderr, "usage: %s [-c chunksize in kb] [-h] "
		"device_name\n", prog);
	exit(1);
}

static int ul_log2(unsigned long arg)
{
        int     l = 0;

        arg >>= 1;
        while (arg) {
                l++;
                arg >>= 1;
        }
        return l;
}

void init_chunk_info(ext2_filsys fs, struct chunk_info *info)
{
	int i;

	info->chunkbits = ul_log2(info->chunkbytes);
	info->blocksize_bits = ul_log2((unsigned long)fs->blocksize);
	info->blks_in_chunk = info->chunkbytes >> info->blocksize_bits;

	info->min = ~0UL;
	info->max = info->avg = 0;
	info->real_free_chunks = 0;

	for (i = 0; i < MAX_HIST; i++)
		info->histogram.fc_buckets[i] = 0;
}

void scan_block_bitmap(ext2_filsys fs, struct chunk_info *info)
{
	unsigned long blocks_count = fs->super->s_blocks_count;
	unsigned long chunks = (blocks_count + info->blks_in_chunk) >>
				(info->chunkbits - info->blocksize_bits);
	unsigned long chunk_num;
	unsigned long last_chunk_size = 0;
	unsigned long index;
	blk_t blk;
	int ret, not_free = 0, free_chunk = 0;

	for (chunk_num = 0; chunk_num < chunks; chunk_num++) {
		blk_t chunk_start_blk = chunk_num << (info->chunkbits -
						      info->blocksize_bits);
		unsigned long num_blks;

		/* Last chunk may be smaller */
		if (chunk_start_blk + info->blks_in_chunk > blocks_count)
			num_blks = blocks_count - chunk_start_blk;
		else
			num_blks = info->blks_in_chunk;

		free_chunk = 0;

		/* Initialize starting block for first chunk correctly else
		 * there is a segfault when blocksize = 1024 in which case
		 * block_map->start = 1 */
		for (blk = (chunk_num == 0 ? fs->super->s_first_data_block : 0);
		     blk < num_blks; blk++) {
			if (ext2fs_fast_test_block_bitmap(fs->block_map,
						 chunk_start_blk + blk)) {
				not_free = 1;
			} else {
				last_chunk_size++;
				free_chunk++;
				not_free = 0;
			}

			if (not_free) {
				if (last_chunk_size == 0)
					continue;

				index = ul_log2(last_chunk_size) + 1;
				info->histogram.fc_buckets[index]++;

				if (last_chunk_size > info->max)
					info->max = last_chunk_size;
				if (last_chunk_size < info->min)
					info->min = last_chunk_size;
				info->avg += last_chunk_size;

				info->real_free_chunks++;
				last_chunk_size = 0;
			}
		}

		if (free_chunk == info->blks_in_chunk)
			info->free_chunks++;
	}
}

errcode_t get_chunk_info(ext2_filsys fs, struct chunk_info *info)
{
	unsigned long total_chunks;
	char *unitp = "KMGTPEZY";
	int units = 10;
	unsigned long start = 0, end, cum;
	int i, retval = 0;

	scan_block_bitmap(fs, info);

	printf("\nTotal blocks: %lu\nFree blocks: %lu (%0.1f%%)\n",
	       fs->super->s_blocks_count, fs->super->s_free_blocks_count,
	       (double)fs->super->s_free_blocks_count * 100 /
						fs->super->s_blocks_count);

	total_chunks = (fs->super->s_blocks_count + info->blks_in_chunk) >>
                                       (info->chunkbits - info->blocksize_bits);
	printf("\nTotal chunks: %lu\nFree chunks: %lu (%0.1f%%)\n",
	       total_chunks, info->free_chunks,
	       (double)info->free_chunks * 100 / total_chunks);

	/* Display chunk information in KB */
	if (info->real_free_chunks) {
		info->min = (info->min * fs->blocksize) >> 10;
		info->max = (info->max * fs->blocksize) >> 10;
		info->avg = (info->avg / info->real_free_chunks *
			     fs->blocksize) >> 10;
	} else {
		info->min = 0;
	}

	printf("\nMin free chunk: %lu KB \nMax free chunk: %lu KB\n"
	       "Avg. free chunk: %lu KB\n", info->min, info->max, info->avg);

	printf("\nHISTOGRAM OF FREE CHUNK SIZES:\n");
	printf("%15s\t\t%10s\n", "Range", "Free chunks");
	for (i = 0; i < MAX_HIST; i++) {
		end = 1 << (i + info->blocksize_bits - units);
		if (info->histogram.fc_buckets[i] != 0)
			printf("%5lu%c...%5lu%c- :  %10lu\n", start, *unitp,
			       end, *unitp, info->histogram.fc_buckets[i]);
		start = end;
		if (start == 1<<10) {
			start = 1;
			units += 10;
			unitp++;
		}
	}

	return retval;
}

void close_device(char *device_name, ext2_filsys fs)
{
	int retval = ext2fs_close(fs);

	if (retval)
		com_err(device_name, retval, "while closing the filesystem.\n");
}

void collect_info(ext2_filsys fs, struct chunk_info *chunk_info)
{
	unsigned int retval = 0, i, free_blks;

	printf("Device: %s\n", fs->device_name);
	printf("Blocksize: %u bytes\n", fs->blocksize);

	retval = ext2fs_read_block_bitmap(fs);
	if (retval) {
		com_err(fs->device_name, retval, "while reading block bitmap");
		close_device(fs->device_name, fs);
		exit(1);
	}

	init_chunk_info(fs, chunk_info);

	retval = get_chunk_info(fs, chunk_info);
	if (retval) {
		com_err(fs->device_name, retval, "while collecting chunk info");
                close_device(fs->device_name, fs);
		exit(1);
	}
}

void open_device(char *device_name, ext2_filsys *fs)
{
	int retval;
	int flag = EXT2_FLAG_FORCE;

	retval = ext2fs_open(device_name, flag, 0, 0, unix_io_manager, fs);
	if (retval) {
		com_err(device_name, retval, "while opening filesystem");
		exit(1);
	}
}

int main(int argc, char *argv[])
{
	struct chunk_info chunk_info = { .chunkbytes = DEFAULT_CHUNKSIZE };
	errcode_t retval = 0;
	ext2_filsys fs = NULL;
	char *device_name;
	char *progname;
	char c, *end;

	progname = argv[0];

	while ((c = getopt(argc, argv, "c:h")) != EOF) {
		switch (c) {
		case 'c':
			chunk_info.chunkbytes = strtoull(optarg, &end, 0);
			if (*end != '\0') {
				fprintf(stderr, "%s: bad chunk size '%s'\n",
					progname, optarg);
				usage(progname);
			}
			if (chunk_info.chunkbytes &
			    (chunk_info.chunkbytes - 1)) {
				fprintf(stderr, "%s: chunk size must be a "
					"power of 2.");
				usage(progname);
			}
			chunk_info.chunkbytes *= 1024;
			break;
		default:
			fprintf(stderr, "%s: bad option '%c'\n",
				progname, c);
		case 'h':
			usage(progname);
			break;
		}
	}

	if (optind == argc) {
		fprintf(stderr, "%s: missing device name.\n", progname);
		usage(progname);
	}

	device_name = argv[optind];

	open_device(device_name, &fs);

	if (chunk_info.chunkbytes < fs->blocksize) {
		fprintf(stderr, "%s: chunksize must be greater than or equal "
			"to filesystem blocksize.\n", progname);
		exit(1);
	}
	collect_info(fs, &chunk_info);
	close_device(device_name, fs);

	return retval;
}
